from checkov.graph.checks.checks_infra.enums import SolverType
from checkov.graph.checks.checks_infra.solvers.base_solver import BaseSolver


class BaseAttributeSolver(BaseSolver):
    operator = ''

    def __init__(self, resource_types, query_attribute, query_value):
        super().__init__(SolverType.ATTRIBUTE)
        self.resource_types = resource_types
        self.query_attribute = query_attribute
        self.query_value = query_value

    def run_query(self, graph_connector):
        # TODO
        raise NotImplementedError

    def get_operation(self, *args):
        # TODO
        raise NotImplementedError

    def get_negative_operation(self, *args):
        # TODO
        raise NotImplementedError

    def _get_operation(self, *args):
        raise NotImplementedError




from checkov.graph.checks.checks_infra.solvers.base_solver import BaseSolver
from checkov.graph.checks.checks_infra.enums import SolverType


class BaseConnectionSolver(BaseSolver):

    operator = ''

    def __init__(self, resource_types, connected_resources_types, vertices_under_resource_types=None, vertices_under_connected_resources_types=None):
        super().__init__(SolverType.CONNECTION)
        self.resource_types = resource_types
        self.connected_resources_types = connected_resources_types
        self.vertices_under_resource_types = vertices_under_resource_types or []
        self.vertices_under_connected_resources_types = vertices_under_connected_resources_types or []

    def run_query(self, graph_connector):
        raise NotImplementedError

    def get_operation(self, *args):
        pass

    def _get_operation(self, *args):
        pass



from checkov.graph.terraform.checks_infra.solvers.attribute_queries import *
from checkov.graph.terraform.checks_infra.solvers.complex_queries import *
from checkov.graph.terraform.checks_infra.solvers.connections_queries import *
from checkov.graph.terraform.checks_infra.solvers.filter_queries import *


import re

from checkov.graph.terraform.utils.utils import INTERPOLATION_PATTERN
from checkov.graph.terraform.variable_rendering.safe_eval_functions import SAFE_EVAL_DICT

# condition ? true_val : false_val -> (condition, true_val, false_val)
CONDITIONAL_EXPR = r'([^\s]+)\?([^\s^\:]+)\:([^\s^\:]+)'

# {key1 = value1, key2 = value2, ...}
MAP_REGEX = r'\{(?:\s*[\S]+\s*\=\s*[\S]+\s*\,)+(?:\s*[\S]+\s*\=\s*[\S]+\s*)\}'

KEY_VALUE_REGEX = r'([\S]+)\s*\=\s*([\S]+)'

# %{ some_text }
DIRECTIVE_EXPR = r'\%\{([^\}]*)\}'

COMPARE_REGEX = re.compile(r'^(?P<a>.+)(?P<operator>==|!=|>=|>|<=|<|&&|\|\|)+(?P<b>.+)$')


def evaluate_terraform(input_str):
    evaluated_value = _try_evaluate(input_str)
    if type(evaluated_value) is not str:
        return input_str if callable(evaluated_value) else evaluated_value
    evaluated_value = evaluate_map(evaluated_value)
    evaluated_value = strip_double_quotes(evaluated_value)
    evaluated_value = evaluate_directives(evaluated_value)
    evaluated_value = evaluate_conditional_expression(evaluated_value)
    evaluated_value = evaluate_compare(evaluated_value)
    second_evaluated_value = _try_evaluate(evaluated_value)

    return evaluated_value if callable(second_evaluated_value) else second_evaluated_value


def _try_evaluate(input_str):
    try:
        return eval(input_str, {"__builtins__": None}, SAFE_EVAL_DICT) # nosec
    except Exception:
        try:
            return eval(f'"{input_str}"', {"__builtins__": None}, SAFE_EVAL_DICT) # nosec
        except Exception:
            return input_str


def replace_string_value(original_str, str_to_replace, replaced_value):
    if type(original_str) is list:
        for i, item in enumerate(original_str):
            original_str[i] = replace_string_value(item, str_to_replace, replaced_value)
            return original_str

    if str_to_replace not in original_str:
        return original_str

    string_without_interpolation = re.sub(INTERPOLATION_PATTERN, '', original_str)
    return string_without_interpolation.replace(str_to_replace, replaced_value)


def strip_double_quotes(input_str):
    if input_str.startswith('"') and input_str.endswith('"'):
        input_str = input_str[1:-1]
    return input_str


def evaluate_conditional_expression(input_str):
    conditions = re.findall(CONDITIONAL_EXPR, re.sub(r'\s', '', input_str))
    for condition in conditions:
        if len(condition) != 3:
            return input_str
        evaluated_condition = evaluate_terraform(condition[0])
        if convert_to_bool(evaluated_condition):
            return evaluate_terraform(condition[1])
        return evaluate_terraform(condition[2])
    return input_str


def evaluate_compare(input_str):
    """
    :param input_str: string like "a && b" (supported operators: ==, != , <, <=, >, >=, && , ||)
    :return: evaluation of the expression
    """
    if type(input_str) is str and re.search(COMPARE_REGEX, input_str) and 'for' not in input_str:
        compare_parts = re.search(COMPARE_REGEX, input_str).groupdict()
        a = compare_parts.get('a')
        b = compare_parts.get('b')
        op = compare_parts.get('operator')
        if a and b and op:
            try:
                return apply_binary_op(evaluate_terraform(a), evaluate_terraform(b), op)
            except TypeError or SyntaxError:
                return input_str

    return input_str


def apply_binary_op(a, b, operator):
    # apply the operator after verifying that a and b have the same type.
    operators = {
        '==': lambda a, b: a == b,
        '!=': lambda a, b: a != b,
        '>': lambda a, b: a > b,
        '>=': lambda a, b: a >= b,
        '<': lambda a, b: a < b,
        '<=': lambda a, b: a <= b,
        '&&': lambda a, b: a and b,
        '||': lambda a, b: a or b
    }
    type_a = type(a)
    type_b = type(b)

    if type_a != type_b:
        try:
            temp_b = type_a(b)
            if type_a == bool:
                temp_b = bool(convert_to_bool(b))
            return operators[operator](a, temp_b)
        except Exception:
            temp_a = type_b(a)
            if type_b == bool:
                temp_a = bool(convert_to_bool(a))
            return operators[operator](temp_a, b)
    else:
        return operators[operator](a, b)


def evaluate_directives(input_str):
    if re.search(DIRECTIVE_EXPR, input_str) is None:
        return input_str

    # replace `%{if <BOOL>}%{true_val}%{else}%{false_val}%{endif}` pattern with `<BOOL> ? true_val : false_val`
    matching_directives = re.findall(DIRECTIVE_EXPR, input_str)
    if len(matching_directives) == 3:
        if re.search(r'\bif\b', matching_directives[0]) and re.search(r'\belse\b', matching_directives[1]) and re.search(r'\bendif\b', matching_directives[2]):
            split_by_directives = re.split(DIRECTIVE_EXPR, input_str)
            edited_str = ''
            for part in split_by_directives:
                if re.search(r'\bif\b', part):
                    part = part.replace('if', '%{') + ' ? '
                    part = re.sub(r'\s', '', part)
                if re.search(r'\belse\b', part):
                    part = part.replace('else', ':')
                    part = re.sub(r'\s', '', part)
                if re.search(r'\bendif\b', part):
                    part = part.replace('endif', '}')
                    part = re.sub(r'\s', '', part)
                edited_str += part
            input_str = edited_str

    matching_directives = re.split(DIRECTIVE_EXPR, input_str)
    evaluated_string_parts = []
    for str_part in matching_directives:
        evaluated_string_parts.append(evaluate_terraform(str_part))
    return ''.join(evaluated_string_parts)


def evaluate_map(input_str):
    matching_maps = re.findall(MAP_REGEX, input_str)
    for matching_map in matching_maps:
        # replaced_matching_map = f'"{matching_map.replace("=", ":")}"'
        replaced_matching_map = matching_map.replace("=", ":")
        input_str = input_str.replace(matching_map, replaced_matching_map)
    return input_str


def convert_to_bool(bool_str):
    if bool_str in ['true', '"true"', 'True', '"True"', 1, '1']:
        return True
    elif bool_str in ['false', '"false"', 'False', '"False"', 0, '0']:
        return False
    else:
        return bool_str



from abc import abstractmethod


class BaseSolver:
    operator = ''

    def __init__(self, query_type):
        self.query_type = query_type

    @abstractmethod
    def get_operation(self, *args):
        raise NotImplementedError()

    @abstractmethod
    def _get_operation(self, *args):
        raise NotImplementedError()

    @abstractmethod
    def run_query(self, graph_connector):
        raise NotImplementedError()




from checkov.graph.checks.checks_infra.solvers.attribute_queries.base_attribute_solver import BaseAttributeSolver


class NotExistsAttributeSolver(BaseAttributeSolver):
    operator = 'not_exists'

    def __init__(self, resource_types, query_attribute, query_value):
        super().__init__(resource_types=resource_types,
                         query_attribute=query_attribute, query_value=query_value)

    def _get_operation(self, *args):
        # TODO
        raise NotImplementedError



from checkov.graph.terraform.checks_infra.solvers.connections_queries.and_connection_solver import AndConnectionSolver
from checkov.graph.terraform.checks_infra.solvers.connections_queries.complex_connection_solver import ComplexConnectionSolver
from checkov.graph.terraform.checks_infra.solvers.connections_queries.connection_exists_solver import ConnectionExistsSolver
from checkov.graph.terraform.checks_infra.solvers.connections_queries.connection_not_exists_solver import ConnectionNotExistsSolver
from checkov.graph.terraform.checks_infra.solvers.connections_queries.or_connection_solver import OrConnectionSolver


from checkov.graph.terraform.checks_infra.solvers.attribute_queries.any_attribute_solver import AnyResourceSolver
from checkov.graph.terraform.checks_infra.solvers.attribute_queries.array_contains_attribute_solver import ArrayContainsAttributeSolver
from checkov.graph.terraform.checks_infra.solvers.attribute_queries.array_not_contains_attribute_solver import ArrayNotContainsAttributeSolver
from checkov.graph.terraform.checks_infra.solvers.attribute_queries.ending_with_attribute_solver import EndingWithAttributeSolver
from checkov.graph.terraform.checks_infra.solvers.attribute_queries.equals_attribute_solver import EqualsAttributeSolver
from checkov.graph.terraform.checks_infra.solvers.attribute_queries.exists_attribute_solver import ExistsAttributeSolver
from checkov.graph.terraform.checks_infra.solvers.attribute_queries.not_ending_with_attribute_solver import NotEndingWithAttributeSolver
from checkov.graph.terraform.checks_infra.solvers.attribute_queries.not_equals_attribute_solver import NotEqualsAttributeSolver
from checkov.graph.terraform.checks_infra.solvers.attribute_queries.not_exists_attribute_solver import NotExistsAttributeSolver
from checkov.graph.terraform.checks_infra.solvers.attribute_queries.not_starting_with_attribute_solver import NotStartingWithAttributeSolver
from checkov.graph.terraform.checks_infra.solvers.attribute_queries.starting_with_attribute_solver import StartingWithAttributeSolver
from checkov.graph.terraform.checks_infra.solvers.attribute_queries.within_attribute_solver import WithinAttributeSolver


from checkov.common.models.enums import CheckResult, CheckCategories
from checkov.arm.base_resource_check import BaseResourceCheck

# https://docs.microsoft.com/en-us/azure/templates/microsoft.storage/storageaccounts/blobservices/containers
# publicAccess default is None

class StorageBlobServiceContainerPrivateAccess(BaseResourceCheck):
    def __init__(self):
        name = "Ensure that 'Public access level' is set to Private for blob containers"
        id = "CKV_AZURE_34"
        supported_resources = ['Microsoft.Storage/storageAccounts/blobServices/containers', 'containers',
                               'blobServices/containers']
        categories = [CheckCategories.NETWORKING]
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources)

    def scan_resource_conf(self, conf):
        if "properties" in conf:
            if "publicAccess" in conf["properties"]:
                if str(conf["properties"]["publicAccess"]).lower() == "container" or \
                        str(conf["properties"]["publicAccess"]).lower() == "blob":
                    return CheckResult.FAILED
        return CheckResult.PASSED

check = StorageBlobServiceContainerPrivateAccess()


import json
import logging
import os

import yaml

from checkov.graph.checks.checks_infra.base_parser import BaseGraphCheckParser
from checkov.graph.checks.checks_infra.registry import BaseRegistry
from checkov.graph.terraform.checks_infra import resources_types

CHECKS_POSSIBLE_ENDING = [".yaml", ".yml"]


class Registry(BaseRegistry):
    def __init__(self, parser=BaseGraphCheckParser()):
        super().__init__(parser)
        self.checks = []
        self.parser = parser

    def load_checks(self):
        checks_dir = os.path.join(os.path.dirname(os.path.dirname(__file__)), "checks")
        for root, d_names, f_names in os.walk(checks_dir):
            for file in f_names:
                file_ending = os.path.splitext(file)[1]
                if file_ending in CHECKS_POSSIBLE_ENDING:
                    with open(f'{root}/{file}', "r") as f:
                        logging.info(f"loading {file}")
                        check_yaml = yaml.safe_load(f)
                        check_json = json.loads(json.dumps(check_yaml))
                        # TODO: implement
                        # check = self.parser.parse_raw_check(check_json, resources_types=self._get_resource_types(check_json))
                        # self.checks.append(check)

    @staticmethod
    def _get_resource_types(check_json):
        provider = check_json.get("scope", {}).get("provider", "").lower()
        return resources_types.get(provider)



from copy import deepcopy

from checkov.graph.terraform.graph_builder.graph_components.block_types import BlockType
from checkov.graph.terraform.graph_builder.graph_components.blocks import Block, get_inner_attributes


class Module:
    def __init__(self, source_dir):
        self.path = ''
        self.blocks = []
        self.customer_name = ''
        self.account_id = ''
        self.source = ''
        self.resources_types = set()
        self.source_dir = source_dir

    def add_blocks(self, block_type, blocks, path, source):
        self.source = source
        if self._block_type_to_func.get(block_type):
            self._block_type_to_func[block_type].__call__(self, blocks, path)

    def _add_provider(self, blocks, path):
        for provider_dict in blocks:
            for name in provider_dict:
                attributes = provider_dict[name]
                provider_name = name
                if attributes.get('alias'):
                    provider_name = f"{provider_name}.{attributes.get('alias')[0]}"
                provider_block = Block(
                    block_type=BlockType.PROVIDER,
                    name=provider_name,
                    config=provider_dict,
                    path=path,
                    attributes=attributes,
                    source=self.source
                )
                self.blocks.append(provider_block)

    def _add_variable(self, blocks, path):
        for variable_dict in blocks:
            for name in variable_dict:
                attributes = variable_dict[name]
                variable_block = Block(
                    block_type=BlockType.VARIABLE,
                    name=name,
                    config=variable_dict,
                    path=path,
                    attributes=attributes,
                    source=self.source
                )
                self.blocks.append(variable_block)

    def _add_locals(self, blocks, path):
        for blocks_section in blocks:
            for name in blocks_section:
                local_block = Block(
                    block_type=BlockType.LOCALS,
                    name=name,
                    config=blocks_section,
                    path=path,
                    attributes={name: blocks_section[name]},
                    source=self.source
                )
                self.blocks.append(local_block)

    def _add_output(self, blocks, path):
        for output_dict in blocks:
            for name in output_dict:
                if type(output_dict[name]) is not dict:
                    continue
                output_block = Block(
                    block_type=BlockType.OUTPUT,
                    name=name,
                    config=output_dict,
                    path=path,
                    attributes={'value': output_dict[name].get('value')},
                    source=self.source
                )
                self.blocks.append(output_block)

    def _add_module(self, blocks, path):
        for module_dict in blocks:
            for name in module_dict:
                module_block = Block(
                    block_type=BlockType.MODULE,
                    name=name,
                    config=module_dict,
                    path=path,
                    attributes=module_dict[name],
                    source=self.source
                )
                self.blocks.append(module_block)

    def _add_resource(self, blocks, path):
        for resource_dict in blocks:
            for resource_type in resource_dict:
                self.resources_types.add(resource_type)
                for name in resource_dict[resource_type]:
                    attributes = deepcopy(resource_dict[resource_type][name])
                    provisioner = attributes.get('provisioner')
                    if provisioner is not None:
                        self._handle_provisioner(provisioner, attributes)
                    attributes['resource_type'] = [resource_type]
                    resource_block = Block(
                        block_type=BlockType.RESOURCE,
                        name=resource_type + '.' + name,
                        config=resource_dict,
                        path=path,
                        attributes=attributes,
                        id=resource_type + '.' + name,
                        source=self.source
                    )
                    self.blocks.append(resource_block)

    def _add_data(self, blocks, path):
        for data_dict in blocks:
            for data_type in data_dict:
                for name in data_dict[data_type]:
                    data_block = Block(
                        block_type=BlockType.DATA,
                        name=data_type + '.' + name,
                        config=data_dict,
                        path=path,
                        attributes=data_dict.get(data_type, {}).get(name, {}),
                        id=data_type + '.' + name,
                        source=self.source
                    )
                    self.blocks.append(data_block)

    def _add_terraform_block(self, blocks, path):
        for terraform_dict in blocks:
            for name in terraform_dict:
                terraform_block = Block(
                    block_type=BlockType.TERRAFORM,
                    name=name,
                    config=terraform_dict,
                    path=path,
                    attributes={},
                    source=self.source
                )
                self.blocks.append(terraform_block)

    def _add_tf_var(self, blocks, path):
        for tf_var_name in blocks:
            tfvar_block = Block(
                block_type=BlockType.TF_VARIABLE,
                name=tf_var_name,
                config=blocks[tf_var_name],
                path=path,
                attributes={tf_var_name: blocks[tf_var_name]},
                source=self.source
            )
            self.blocks.append(tfvar_block)

    @staticmethod
    def _handle_provisioner(provisioner, attributes):
        for pro in provisioner:
            if pro.get('local-exec'):
                inner_attributes = get_inner_attributes('provisioner/local-exec', pro['local-exec'])
                attributes.update(inner_attributes)
            elif pro.get('remote-exec'):
                inner_attributes = get_inner_attributes('provisioner/remote-exec', pro['remote-exec'])
                attributes.update(inner_attributes)
        attributes.pop('provisioner')

    def get_resources_types(self):
        return list(self.resources_types)

    _block_type_to_func = {
        BlockType.DATA: _add_data,
        BlockType.LOCALS: _add_locals,
        BlockType.MODULE: _add_module,
        BlockType.OUTPUT: _add_output,
        BlockType.PROVIDER: _add_provider,
        BlockType.RESOURCE: _add_resource,
        BlockType.TERRAFORM: _add_terraform_block,
        BlockType.TF_VARIABLE: _add_tf_var,
        BlockType.VARIABLE: _add_variable,
    }



from dataclasses import dataclass

from checkov.graph.graph_builder.graph_components.attribute_names import CustomAttributes, props


@dataclass
class CustomAttributes(CustomAttributes):
    SOURCE_MODULE = "source_module_"
    ENCRYPTION = "encryption_"
    ENCRYPTION_DETAILS = "encryption_details_"


reserved_attribute_names = props(CustomAttributes)



from checkov.graph.checks.checks_infra.solvers.connection_queries.base_connection_solver import BaseConnectionSolver


class ConnectionExistsSolver(BaseConnectionSolver):
    operator = 'exists'

    def __init__(self, resource_types, connected_resources_types, vertices_under_resource_types=None, vertices_under_connected_resources_types=None):
        super().__init__(resource_types, connected_resources_types, vertices_under_resource_types, vertices_under_connected_resources_types)

    def run_query(self, graph_connector):
        # TODO
        raise NotImplementedError



import logging
import os
from copy import deepcopy

from checkov.graph.terraform.graph_builder.graph_components.attribute_names import CustomAttributes
from checkov.graph.terraform.graph_builder.graph_components.block_types import BlockType
from checkov.graph.terraform.utils.utils import get_referenced_vertices_in_value, run_function_multithreaded, calculate_hash, \
    join_trimmed_strings, remove_interpolation, remove_index_pattern_from_str, remove_function_calls_from_str
from checkov.graph.terraform.variable_rendering.evaluate_terraform import replace_string_value, evaluate_terraform


class VariableRenderer:
    def __init__(self, local_graph):
        self.local_graph = local_graph
        run_async = os.environ.get('RENDER_VARIABLES_ASYNC')
        if not run_async:
            run_async = True
        else:
            run_async = True if run_async == 'True' else False
        self.run_async = run_async
        max_workers = os.environ.get('RENDER_ASYNC_MAX_WORKERS')
        if not max_workers:
            max_workers = 50
        else:
            max_workers = int(max_workers)
        self.max_workers = max_workers
        self.done_edges = []

    def render_variables_from_local_graph(self):
        # find vertices with out-degree = 0 and in-degree > 0
        end_vertices_indexes = self.local_graph.get_vertices_with_degrees_conditions(
            out_degree_cond=lambda degree: degree == 0,
            in_degree_cond=lambda degree: degree > 0)

        # all the edges entering `end_vertices`
        edges_to_render = self.local_graph.get_in_edges(end_vertices_indexes)
        while len(edges_to_render) > 0:
            logging.info(f'evaluating {len(edges_to_render)} edges')
            # group edges that have the same origin and label together
            edges_groups = self.group_edges_by_origin_and_label(edges_to_render)
            if self.run_async:
                run_function_multithreaded(func=self._edge_evaluation_task, data=edges_groups, max_group_size=1,
                                           num_of_workers=self.max_workers)
            else:
                for edge_group in edges_groups:
                    self._edge_evaluation_task([edge_group])
            end_vertices_indexes = list(set([edge.origin for edge in edges_to_render]))
            edges_to_render = self.local_graph.get_in_edges(end_vertices_indexes)
            edges_to_render = [edge for edge in edges_to_render if edge not in self.done_edges]
        self.local_graph.update_vertices_configs()
        logging.info('done evaluating edges')

    def _edge_evaluation_task(self, edges):
        edges = edges[0]
        self.evaluate_vertex_attribute_from_edge(edges)
        return edges

    def evaluate_vertex_attribute_from_edge(self, edge_list):
        multiple_edges = len(edge_list) > 1
        edge = edge_list[0]
        origin_vertex_attributes = self.local_graph.get_vertex_attributes_by_index(edge.origin)
        val_to_eval = deepcopy(origin_vertex_attributes.get(edge.label, ''))

        referenced_vertices = get_referenced_vertices_in_value(value=val_to_eval, aliases={},
                                                               resources_types=self.local_graph.get_resources_types_in_graph(),
                                                               cleanup_functions=[remove_function_calls_from_str,
                                                                                  remove_interpolation])
        first_key_path = None
        if referenced_vertices:
            for edge in edge_list:
                dest_vertex_attributes = self.local_graph.get_vertex_attributes_by_index(edge.dest)
                key_path_in_dest_vertex, replaced_key = self.find_path_from_referenced_vertices(referenced_vertices,
                                                                                                  dest_vertex_attributes)
                if not key_path_in_dest_vertex or not replaced_key:
                    continue
                if not first_key_path:
                    first_key_path = key_path_in_dest_vertex

                evaluated_attribute_value = self.extract_value_from_vertex(key_path_in_dest_vertex,
                                                                           dest_vertex_attributes)
                if evaluated_attribute_value is not None:
                    val_to_eval = replace_string_value(original_str=val_to_eval, str_to_replace=replaced_key, replaced_value=str(evaluated_attribute_value))
                if not multiple_edges:
                    self.update_evaluated_value(changed_attribute_key=edge.label,
                                                changed_attribute_value=val_to_eval, vertex=edge.origin, change_origin_id=edge.dest, attribute_at_dest=key_path_in_dest_vertex)
        if multiple_edges:
            self.update_evaluated_value(changed_attribute_key=edge.label,
                                        changed_attribute_value=val_to_eval, vertex=edge.origin, change_origin_id=edge.dest, attribute_at_dest=first_key_path)

    @staticmethod
    def extract_value_from_vertex(key_path, attributes):

        for i in range(len(key_path)):
            key = join_trimmed_strings(char_to_join=".", str_lst=key_path, num_to_trim=i)
            value = attributes.get(key, None)
            if value is not None:
                return value

        reversed_key_path = deepcopy(key_path)
        reversed_key_path.reverse()
        for i in range(len(reversed_key_path)):
            key = join_trimmed_strings(char_to_join=".", str_lst=reversed_key_path, num_to_trim=i)
            value = attributes.get(key, None)
            if value is not None:
                return value

        if attributes.get(CustomAttributes.BLOCK_TYPE) == BlockType.VARIABLE.value:
            return attributes.get('default')
        if attributes.get(CustomAttributes.BLOCK_TYPE) == BlockType.OUTPUT.value:
            return attributes.get('value')
        return None

    @staticmethod
    def find_path_from_referenced_vertices(referenced_vertices, vertex_attributes):
        """
        :param referenced_vertices: an array of VertexReference
        :param vertex_attributes: attributes to search
        :return attribute_path: [] if referenced_vertices does not contain vertex_attributes,
                                else the path to the searched attribute: ['vpc_id']
        :return origin_value
        """
        for vertex_reference in referenced_vertices:
            block_type = vertex_reference.block_type.value
            attribute_path = vertex_reference.sub_parts
            copy_of_attribute_path = deepcopy(attribute_path)
            if vertex_attributes[CustomAttributes.BLOCK_TYPE] == block_type:
                for i in range(len(copy_of_attribute_path)):
                    copy_of_attribute_path[i] = remove_index_pattern_from_str(copy_of_attribute_path[i])
                    name = ".".join(copy_of_attribute_path[:i + 1])
                    if vertex_attributes[CustomAttributes.BLOCK_NAME] == name:
                        return attribute_path, vertex_reference.origin_value
            elif block_type == BlockType.MODULE:
                copy_of_attribute_path.reverse()
                for i in range(len(copy_of_attribute_path)):
                    copy_of_attribute_path[i] = remove_index_pattern_from_str(copy_of_attribute_path[i])
                    name = ".".join(copy_of_attribute_path[:i + 1])
                    if vertex_attributes[CustomAttributes.BLOCK_NAME] == name:
                        return name.split('.'), vertex_reference.origin_value
        return [], ''

    def update_evaluated_value(self, changed_attribute_key, changed_attribute_value, vertex, change_origin_id, attribute_at_dest=None):
        """
        The function updates the value of changed_attribute_key with changed_attribute_value for vertex
        """
        references_vertices = get_referenced_vertices_in_value(value=changed_attribute_value, aliases={},
                                                               resources_types=self.local_graph.get_resources_types_in_graph())
        if not references_vertices:
            evaluated_attribute_value = evaluate_terraform(f'"{str(changed_attribute_value)}"')
            self.local_graph.update_vertex_attribute(vertex, changed_attribute_key, evaluated_attribute_value, change_origin_id, attribute_at_dest)

    @staticmethod
    def group_edges_by_origin_and_label(edges):
        edge_groups = {}
        for edge in edges:
            origin_and_label_hash = calculate_hash(f'{edge.origin}{edge.label}')
            if not edge_groups.get(origin_and_label_hash):
                edge_groups[origin_and_label_hash] = []
            edge_groups[origin_and_label_hash].append(edge)
        return list(edge_groups.values())



from checkov.graph.checks.checks_infra.solvers.connection_queries.base_connection_solver import BaseConnectionSolver
from checkov.graph.checks.checks_infra.enums import SolverType


class ComplexConnectionSolver(BaseConnectionSolver):
    operator = ''

    def __init__(self, queries, operator):
        if queries is None:
            queries = []
        self.queries = queries
        self.operator = operator

        resource_types = []
        connected_resources_types = []
        for sub_query in self.queries:
            if sub_query.query_type in [SolverType.CONNECTION, SolverType.COMPLEX_CONNECTION]:
                resource_types.extend(sub_query.resource_types)
                connected_resources_types.extend(sub_query.connected_resources_types)
        resource_types = list(set(resource_types))
        connected_resources_types = list(set(connected_resources_types))

        super().__init__(resource_types, connected_resources_types)

    def run_query(self, graph_connector):
        # TODO
        raise NotImplementedError

    def get_operation(self, *args):
        raise NotImplementedError

    def filter_results(self, traversal):
        raise NotImplementedError



import base64
import builtins
import hashlib
import io
import json
import os
import re
import pickle # nosec

MODULE_DEPENDENCY_PATTERN_IN_PATH = r'\[.+\#.+\]'


def is_local_path(root_dir, source):
    # https://www.terraform.io/docs/modules/sources.html#local-paths
    return source.startswith('./') or \
            source.startswith('/./') or \
            source.startswith('../') or \
            source in os.listdir(root_dir)


def remove_module_dependency_in_path(path):
    """
    :param path: path that looks like "dir/main.tf[other_dir/x.tf#0]
    :return: separated path from module dependency: dir/main.tf, other_dir/x.tf
    """
    module_dependency = re.findall(MODULE_DEPENDENCY_PATTERN_IN_PATH, path)
    if re.findall(MODULE_DEPENDENCY_PATTERN_IN_PATH, path):
        path = re.sub(MODULE_DEPENDENCY_PATTERN_IN_PATH, '', path)
    return path, extract_module_dependency_path(module_dependency)


def extract_module_dependency_path(module_dependency):
    """
    :param module_dependency: a list looking like ['[path_to_module.tf#0]']
    :return: the path without enclosing array and index: 'path_to_module.tf'
    """
    if not module_dependency:
        return ''
    if type(module_dependency) is list and len(module_dependency) > 0:
        module_dependency = module_dependency[0]
    return module_dependency[1:-1].split('#')[0]


def encode_graph_property_value(value, encode_values=True):
    if not encode_values:
        return value
    if isinstance(value, bool):
        # Encode boolean into Terraform's lower case convention
        value = str(value).lower()
    elif isinstance(value, (float, int)):
        value = str(value)
    return json.dumps(value, cls=PythonObjectEncoder, indent=4)


class RestrictedUnpickler(pickle.Unpickler):
    def find_class(self, module, name):
        # Only allow safe classes from builtins.
        if module == "builtins" or module.startswith("lark."):
            return getattr(builtins, name)
        # Forbid everything else.
        raise pickle.UnpicklingError("global '%s.%s' is forbidden" %
                                     (module, name))


def restricted_loads(s):
    """Helper function analogous to pickle.loads()."""
    return RestrictedUnpickler(io.BytesIO(s)).load()


class PythonObjectEncoder(json.JSONEncoder):
    def default(self, obj):
        return {'_python_object':
                base64.b64encode(pickle.dumps(obj)).decode('utf-8') }


def as_python_object(dct):
    if '_python_object' in dct:
        return restricted_loads(base64.b64decode(dct['_python_object']))
    return dct


def calculate_hash(data):
    encoded_attributes = encode_graph_property_value(data)
    sha256 = hashlib.sha256()
    sha256.update(repr(encoded_attributes).encode('utf-8'))

    return sha256.hexdigest()


from checkov.common.output.record import Record


class GraphRecord(Record):
    breadcrumbs = {}

    def __init__(self, record, breadcrumbs):
        self.check_id = record.check_id
        self.check_name = record.check_name
        self.check_result = record.check_result
        self.code_block = record.code_block
        self.file_path = record.file_path
        self.file_line_range = record.file_line_range
        self.resource = record.resource
        self.evaluations = record.evaluations
        self.check_class = record.check_class
        self.repo_file_path = record.repo_file_path
        self.fixed_definition = record.fixed_definition
        self.entity_tags = record.entity_tags
        self.caller_file_path = record.caller_file_path
        self.caller_file_line_range = record.caller_file_line_range
        self.breadcrumbs = breadcrumbs



import copy
import logging
import os
from typing import Dict

import dpath.util

from checkov.common.output.record import Record
from checkov.common.output.report import Report
from checkov.common.runners.base_runner import BaseRunner
from checkov.common.variables.context import EvaluationContext
from checkov.graph.db_connectors.networkx.networkx_db_connector import NetworkxConnector
from checkov.graph.graph_record import GraphRecord
from checkov.graph.terraform.parser import TerraformGraphParser
from checkov.graph.terraform.checks_infra.nx_checks_parser import NXGraphCheckParser
from checkov.graph.terraform.checks_infra.registry import Registry
from checkov.graph.terraform.graph_builder.graph_components.attribute_names import CustomAttributes
from checkov.graph.terraform.graph_builder.graph_to_tf_definitions import convert_graph_vertices_to_tf_definitions
from checkov.graph.terraform.graph_builder.local_graph import LocalGraph
from checkov.graph.terraform.graph_manager import GraphManager
from checkov.runner_filter import RunnerFilter
from checkov.terraform.checks.resource.registry import resource_registry
from checkov.terraform.runner import Runner as TerraformRunner

LOG_LEVEL = os.getenv('LOG_LEVEL', 'WARNING').upper()
logging.basicConfig(level=LOG_LEVEL)


class Runner(BaseRunner):
    check_type = "terraform"

    def __init__(self, parser=TerraformGraphParser(), db_connector=NetworkxConnector()):
        self.parser = parser
        self.tf_definitions = {}
        self.definitions_context = {}
        self.evaluations_context: Dict[str, Dict[str, EvaluationContext]] = {}
        self.graph_manager = GraphManager(source="Terraform", db_connector=db_connector)
        self.tf_runner = TerraformRunner()
        self.graph = None

    def run(self, root_folder, external_checks_dir=None, files=None, runner_filter=RunnerFilter(),
            collect_skip_comments=True, local_graph_class=LocalGraph):
        report = Report(self.check_type)
        self.tf_definitions = {}
        parsing_errors = {}
        breadcrumbs = {}
        if external_checks_dir:
            for directory in external_checks_dir:
                resource_registry.load_external_checks(directory, runner_filter)
        if root_folder:
            root_folder = os.path.abspath(root_folder)

            local_graph, tf_definitions = self.graph_manager.build_graph_from_source_directory(root_folder, local_graph_class)
            self.graph = self.graph_manager.save_graph(local_graph)
            self.tf_runner.tf_definitions, breadcrumbs = convert_graph_vertices_to_tf_definitions(local_graph.vertices,
                                                                                                  root_folder)
            self.tf_runner.check_tf_definition(report, root_folder, runner_filter, collect_skip_comments)

        if files:
            # TODO: support parsing a specific file!!
            files = [os.path.abspath(file) for file in files]
            root_folder = os.path.split(os.path.commonprefix(files))[0]
            for file in files:
                if file.endswith(".tf"):
                    file_parsing_errors = {}
                    parse_result = self.parser.parse_file(file=file, parsing_errors=file_parsing_errors)
                    if parse_result is not None:
                        self.tf_runner.tf_definitions[file] = parse_result
                    if file_parsing_errors:
                        parsing_errors.update(file_parsing_errors)
                        continue
                    self.tf_runner.check_tf_definition(report, root_folder, runner_filter, collect_skip_comments)

        report.add_parsing_errors(parsing_errors.keys())

        graph_report = self.get_graph_checks_report(root_folder, breadcrumbs)
        merge_reports(report, graph_report)

        return report

    def get_graph_checks_report(self, root_folder, breadcrumbs):
        registry = Registry(parser=NXGraphCheckParser())
        registry.load_checks()
        report = Report(self.check_type)
        checks_results = registry.run_checks(self.graph)
        for check_id, check_results in checks_results.items():
            for check_result in check_results:
                entity = check_result['entity']
                entity_context, entity_evaluations = self.get_entity_context_and_evaluations(entity)
                if entity_context:
                    full_file_path = entity[CustomAttributes.FILE_PATH]
                    copy_of_check_result = copy.deepcopy(check_result)
                    copy_of_check_result['entity'] = entity.get(CustomAttributes.CONFIG)
                    record = Record(check_id=check_id,
                                    check_name=check_id,
                                    check_result=copy_of_check_result,
                                    code_block=entity_context.get('code_lines'),
                                    file_path=f"/{os.path.relpath(full_file_path, root_folder)}",
                                    file_line_range=[entity_context.get('start_line'), entity_context.get('end_line')],
                                    resource=".".join(entity_context['definition_path']),
                                    evaluations=entity_evaluations,
                                    check_class=check_id,
                                    file_abs_path=os.path.abspath(full_file_path))
                    breadcrumb = breadcrumbs.get(record.file_path, {}).get(record.resource)
                    if breadcrumb:
                        record = GraphRecord(record, breadcrumb)

                    report.add_record(record=record)
        return report

    def get_entity_context_and_evaluations(self, entity):
        entity_evaluations = None
        block_type = entity[CustomAttributes.BLOCK_TYPE]
        full_file_path = entity[CustomAttributes.FILE_PATH]
        definition_path = entity[CustomAttributes.BLOCK_NAME].split('.')
        entity_context = None
        entity_context_path = [block_type] + definition_path
        if dpath.search(self.tf_runner.definitions_context.get(full_file_path), entity_context_path):
            entity_context = dpath.get(self.tf_runner.definitions_context[full_file_path],
                                       entity_context_path)
            entity_context['definition_path'] = definition_path
        return entity_context, entity_evaluations


def merge_reports(base_report, report_to_merge):
    base_report.passed_checks.extend(report_to_merge.passed_checks)
    base_report.failed_checks.extend(report_to_merge.failed_checks)
    base_report.skipped_checks.extend(report_to_merge.skipped_checks)
    base_report.parsing_errors.extend(report_to_merge.parsing_errors)



from checkov.graph.checks.checks_infra.enums import SolverType
from checkov.graph.checks.checks_infra.solvers.base_solver import BaseSolver


class BaseComplexSolver(BaseSolver):
    operator = ''

    def __init__(self, queries, resource_types):
        if queries is None:
            queries = []
        self.queries = queries
        self.resource_types = resource_types
        super().__init__(SolverType.COMPLEX)

    def run_query(self, graph_connector):
        # TODO
        raise NotImplementedError

    def get_operation(self, *args):
        # TODO
        raise NotImplementedError

    def _get_operation(self, *args):
        pass



from checkov.graph.checks.checks_infra.solvers.complex_queries.base_complex_query import BaseComplexSolver


class AndSolver(BaseComplexSolver):
    operator = 'and'

    def __init__(self, queries, resource_types):
        super().__init__(queries, resource_types)

    def _get_operation(self, *args):
        # TODO
        raise NotImplementedError



from checkov.graph.checks.checks_infra.base_check import BaseGraphCheck
from checkov.graph.checks.checks_infra.base_parser import BaseGraphCheckParser
from checkov.graph.checks.checks_infra.enums import SolverType
from checkov.graph.terraform.checks_infra.solvers import *

operators_to_attributes_query_classes = {
    'equals': EqualsAttributeSolver,
    'not_equals': NotEqualsAttributeSolver,
    'exists': ExistsAttributeSolver,
    'any': AnyResourceSolver,
    'contains': ArrayContainsAttributeSolver,
    'not_exists': NotExistsAttributeSolver,
    'within': WithinAttributeSolver,
    'not_contains': ArrayNotContainsAttributeSolver,
    'starting_with': StartingWithAttributeSolver,
    'not_starting_with': NotStartingWithAttributeSolver,
    'ending_with': EndingWithAttributeSolver,
    'not_ending_with': NotEndingWithAttributeSolver
}

operators_to_complex_query_classes = {
    'and': AndSolver,
    'or': OrSolver,
}

operator_to_connection_query_classes = {
    'exists': ConnectionExistsSolver,
    'not_exists': ConnectionNotExistsSolver
}

operator_to_complex_connection_query_classes = {
    'and': AndConnectionSolver,
    'or': OrConnectionSolver
}

operator_to_filter_query_classes = {
    'within': WithinFilterSolver,
}

condition_type_to_query_type = {
    '': SolverType.ATTRIBUTE,
    'attribute': SolverType.ATTRIBUTE,
    'connection': SolverType.CONNECTION,
    'filter': SolverType.FILTER
}


class NXGraphCheckParser(BaseGraphCheckParser):
    def parse_raw_check(self, raw_check, **kwargs):
        policy_query = raw_check.get("query")
        check = self._parse_raw_check(policy_query, kwargs.get("resources_types"))
        check.id = raw_check.get("metadata", {}).get("id")
        solver = self.get_check_solver(check)
        check.set_solver(solver)

    def _parse_raw_check(self, raw_check, resources_types):
        check = BaseGraphCheck()
        complex_operator = get_complex_operator(raw_check)
        if complex_operator:
            check.type = SolverType.COMPLEX
            check.operator = complex_operator
            sub_queries = raw_check.get(complex_operator)
            for sub_query in sub_queries:
                check.sub_checks.append(self.parse_raw_check(sub_query))
            resources_types_of_sub_queries = [q.resource_types for q in check.sub_checks]
            check.resource_types = list(set(sum(resources_types_of_sub_queries, [])))
            if any(q.type in [SolverType.CONNECTION, SolverType.COMPLEX_CONNECTION] for q in check.sub_checks):
                check.type = SolverType.COMPLEX_CONNECTION

        else:
            resource_type = raw_check.get("resource_types", [])
            if resource_type == ['All'] or resource_type == 'all' or not resource_type:
                check.resource_types = resources_types
            else:
                check.resource_types = resource_type

            connected_resources_type = raw_check.get('connected_resource_types', [])
            if connected_resources_type == ['All'] or connected_resources_type == 'all':
                check.connected_resources_types = resources_types
            else:
                check.connected_resources_types = connected_resources_type

            condition_type = raw_check.get('cond_type', '')
            check.type = condition_type_to_query_type.get(condition_type)
            if condition_type == '':
                check.operator = 'any'
            else:
                check.operator = raw_check.get('operator')
            check.attribute = raw_check.get('attribute')
            check.attribute_value = raw_check.get('value')

        return check

    def get_check_solver(self, check):
        sub_queries_solvers = []
        if check.sub_checks:
            sub_queries_solvers = []
            for sub_query in check.sub_checks:
                sub_queries_solvers.append(self.get_check_solver(sub_query))

        type_to_solver = {
            SolverType.COMPLEX_CONNECTION: operator_to_complex_connection_query_classes.get(check.operator, lambda *args: None)(sub_queries_solvers, check.operator),
            SolverType.COMPLEX: operators_to_complex_query_classes.get(check.operator, lambda *args: None)(sub_queries_solvers, check.resource_types),
            SolverType.ATTRIBUTE: operators_to_attributes_query_classes.get(check.operator, lambda *args: None)(check.resource_types, check.attribute, check.attribute_value),
            SolverType.CONNECTION: operator_to_connection_query_classes.get(check.operator, lambda *args: None)(check.resource_types, check.connected_resources_types),
            SolverType.FILTER: operator_to_filter_query_classes.get(check.operator, lambda *args: None)(check.resource_types, check.attribute, check.attribute_value)
        }

        solver = type_to_solver.get(check.type)
        if not solver:
            raise NotImplementedError(f"query type {check.type} with operator {check.operator} is not supported")
        return solver


def get_complex_operator(raw_check):
    for operator in operators_to_complex_query_classes.keys():
        if raw_check.get(operator):
            return operator
    return None



from checkov.graph.terraform.checks_infra.solvers.connections_queries.complex_connection_solver import ComplexConnectionSolver


class AndConnectionSolver(ComplexConnectionSolver):
    operator = 'and'

    def __init__(self, queries, operator):
        super().__init__(queries, operator)

    def get_operation(self, *args):
        # TODO
        raise NotImplementedError

    def filter_results(self, traversal):
        # TODO
        raise NotImplementedError



from checkov.common.models.enums import CheckCategories, CheckResult
from checkov.terraform.checks.resource.base_resource_negative_value_check import BaseResourceNegativeValueCheck


class ElasticCacheAutomaticBackup(BaseResourceNegativeValueCheck):
    def __init__(self):
        name = "Ensure that Amazon ElastiCache Redis clusters have automatic backup turned on"
        id = "CKV_AWS_112"
        supported_resources = ['aws_elasticache_cluster']
        categories = [CheckCategories.BACKUP_AND_RECOVERY]
        super().__init__(name=name, id=id, categories=categories, supported_resources=supported_resources,
                         missing_attribute_result=CheckResult.FAILED)

    def get_inspected_key(self):
        return 'snapshot_retention_limit'

    def get_forbidden_values(self):
        return [0]


check = ElasticCacheAutomaticBackup()



from checkov.common.models.enums import CheckCategories, CheckResult
from checkov.kubernetes.base_spec_check import BaseK8Check


class SharedHostNetworkNamespace(BaseK8Check):

    def __init__(self):
        # CIS-1.3 1.7.4
        # CIS-1.5 5.2.4
        name = "Containers should not share the host network namespace"
        id = "CKV_K8S_19"
        # Location: Pod.spec.hostNetwork
        # Location: CronJob.spec.jobTemplate.spec.template.spec.hostNetwork
        # Location: *.spec.template.spec.hostNetwork
        supported_kind = ['Pod', 'Deployment', 'DaemonSet', 'StatefulSet', 'ReplicaSet', 'ReplicationController', 'Job', 'CronJob']
        categories = [CheckCategories.KUBERNETES]
        super().__init__(name=name, id=id, categories=categories, supported_entities=supported_kind)

    def get_resource_id(self, conf):
        if "namespace" in conf["metadata"]:
            return "{}.{}.{}".format(conf["kind"], conf["metadata"]["name"], conf["metadata"]["namespace"])
        else:
            return "{}.{}.default".format(conf["kind"], conf["metadata"]["name"])

    def scan_spec_conf(self, conf):
        spec = {}

        if conf['kind'] == 'Pod':
            if "spec" in conf:
                spec = conf["spec"]
        elif conf['kind'] == 'CronJob':
            if "spec" in conf:
                if "jobTemplate" in conf["spec"]:
                    if "spec" in conf["spec"]["jobTemplate"]:
                        if "template" in conf["spec"]["jobTemplate"]["spec"]:
                            if "spec" in conf["spec"]["jobTemplate"]["spec"]["template"]:
                                spec = conf["spec"]["jobTemplate"]["spec"]["template"]["spec"]
        else:
            if "spec" in conf:
                if "template" in conf["spec"]:
                    if "spec" in conf["spec"]["template"]:
                        spec = conf["spec"]["template"]["spec"]
        if spec:
            if "hostNetwork" in spec:
                if spec["hostNetwork"]:
                    return CheckResult.FAILED
                else:
                    return CheckResult.PASSED
            return CheckResult.PASSED
        return CheckResult.FAILED

check = SharedHostNetworkNamespace()



import logging
import os

from checkov.graph.terraform.graph_builder.graph_components.blocks import Block


def convert_graph_vertices_to_tf_definitions(vertices, root_folder):
    tf_definitions = {}
    breadcrumbs = {}
    for vertex in vertices:
        block_path = vertex.path
        if not os.path.isfile(block_path):
            print(f'tried to convert vertex to tf_definitions but its path doesnt exist: {vertex}')
            continue

        block_type = vertex.block_type.value
        if tf_definitions.get(block_path) is None:
            tf_definitions[block_path] = {}
        if tf_definitions.get(block_path).get(block_type) is None:
            tf_definitions.get(block_path)[block_type] = []
        tf_definitions.get(block_path)[block_type].append(vertex.config)
        relative_block_path = f"/{os.path.relpath(block_path, root_folder)}"
        add_breadcrumbs(vertex, breadcrumbs, relative_block_path)
    logging.info(f'breadcrumbs = {breadcrumbs}')
    return tf_definitions, breadcrumbs


def add_breadcrumbs(vertex: Block, breadcrumbs, relative_block_path):
    vertex_breadcrumbs = vertex.breadcrumbs
    if vertex_breadcrumbs:
        if breadcrumbs.get(relative_block_path) is None:
            breadcrumbs[relative_block_path] = {}
        breadcrumbs[relative_block_path][vertex.name] = vertex_breadcrumbs



from checkov.common.models.enums import CheckCategories, CheckResult
from checkov.kubernetes.base_spec_check import BaseK8Check

class ApiServerServiceAccountLookup(BaseK8Check):
    def __init__(self):
        # CIS-1.6 1.2.27
        id = "CKV_K8S_96"
        name = "Ensure that the --service-account-lookup argument is set to true"
        categories = [CheckCategories.KUBERNETES]
        supported_entities = ['containers']
        super().__init__(name=name, id=id, categories=categories, supported_entities=supported_entities)

    def get_resource_id(self, conf):
        return f'{conf["parent"]} - {conf["name"]}'

    def scan_spec_conf(self, conf):
        if "command" in conf:
            if "kube-apiserver" in conf["command"]:
                if "--service-account-lookup=false" in conf["command"] or "--service-account-lookup=true" not in conf["command"]:
                    return CheckResult.FAILED

           
        return CheckResult.PASSED

check = ApiServerServiceAccountLookup()


from checkov.graph.checks.checks_infra.solvers.filter_queries.base_filter_solver import BaseFilterSolver
from checkov.graph.terraform.utils.utils import decode_graph_property_value, encode_graph_property_value


class WithinFilterSolver(BaseFilterSolver):
    operator = 'within'

    def __init__(self, resource_types, query_attribute, query_value):
        query_value = decode_graph_property_value(query_value)
        query_value = [encode_graph_property_value(val) for val in query_value]
        super().__init__(resource_types=resource_types,
                         query_attribute=query_attribute, query_value=query_value)

    def get_operation(self, *args):
        # TODO
        raise NotImplementedError
